########################################
# Kohonen maps to map multi-label data #
# Ricardo Cerri 02/07/2019			       #
########################################


# Search for the initial number of neurons that make it possible to execute kohonen without error
# -------------------------------------------------------------------------------------
getInitialNeuronGrid <- function(mldrDataTrain,novel.classes,num.it,init.n,final.n,input.neurons.grid,tr.mode){
  
  cat("\nGetting the possible initial number of neurons for each class.\n")
  
  # Will store the number of neurons
  n.neurons.grid <- vector(length=length(input.neurons.grid), mode="numeric")
  
  # Parameters of the algorithm
  input.neurons.grid <- input.neurons.grid # the configuration inserted by the user
  n0 <- init.n # initial learning rate
  n1 <- final.n # final learnin rate
  # Learning rate will decline linearly from n0 to n1 over num.iterations updates 
  num.iterations <- num.it # maximum number of iteractions. 
  train.mode <- tr.mode # training mode

  # We cannot allow only two instances in a neuron! That generates an error!
  less.than.three <- 0
  
  # Get 60% of the data for training
  ds <- create_holdout_partition(mldrDataTrain, c(train=0.60, test=0.40), "iterative")
  
  # Dataset, attributes and labels indexes
  mldrDataTrain.labels.index <- ds$train$labels$index
  mldrDataTrain.dataset <- ds$train$dataset
  mldrDataTrain.attributesIndexes <- ds$train$attributesIndexes
  
  # Check for novel classes
  if(novel.classes[1] != 0){
    mldrDataTrain.labels.index <- mldrDataTrain.labels.index[-novel.classes]
  }
  
  # number of initial classes.
  num.classes <- length(mldrDataTrain.labels.index)  

  # Separate data from classes
  data.train <- mldrDataTrain.dataset[,mldrDataTrain.attributesIndexes]
  classes.train <- mldrDataTrain.dataset[,mldrDataTrain.labels.index]
  data.train <- as.matrix(data.train)
  
  # Will train a Kohonen map for each class
  for(i in 1:num.classes){
    
    executed <- 0
    n.neurons <- input.neurons.grid[i]+1
    
    # Execute
    while(executed == 0){
      
      executed <- 1
      n.neurons <- n.neurons - 1
      n.neurons.grid[i] <- n.neurons
      
      # Execute SOM in data from class i
      data = data.train[which(classes.train[,i]==1),]
      
      # Grid dimension = n.neurons x n.neurons
      mygrid <- somgrid(n.neurons,n.neurons, topo=c("hexagonal"), neighbourhood.fct = c("gaussian"))
      res <- try(som.map <- som(data, grid=mygrid, rlen = num.iterations, alpha = c(n0, n1), dist.fcts=c("euclidean"), 
                                mode = c(train.mode), maxNA.fraction = 0L, cores=-1), silent=TRUE)#, keep.data = FALSE)
      
      # Verify if execution was successful 
      if(inherits(res, "try-error")){
        executed <- 0
      }  
    }
  }
  
  return(n.neurons.grid)
}


# -------------------------------------------------------------------------------------


# Search for the best number of neurons in the training data
# -------------------------------------------------------------------------------------
searchNumberOfNeurons <- function(mldrDataTrain,num.iterations,n0,n1,n.neurons.grid,train.mode,min.ex,novel.classes,
                                  num.classes,num.evaluation.windows){
  
  cat("\nSearching for the best number of neurons!\n")
  
  # Parameters of the algorithm
  n0 <- n0 # initial learning rate
  n1 <- n1 # final learnin rate
  # Learning rate will decline linearly from n0 to n1 over num.iterations updates 
  num.iterations <- num.iterations # maximum number of iteractions. 
  train.mode <- train.mode # training mode
  min.ex <- min.ex # minimum number of instances in a neuron
  
  # Number of neurons for the map of each class
  n.neurons <- n.neurons.grid
  
  # Store fmeasure for all configurations
  fmeasures.configurations <- vector()
  
  # Store percentages for all configurations
  percentages.configurations <- vector()
  
  # Store all configurations 
  configurations <- list()
  
  # Number of configurations tested
  n.configurations <- 0
  
  # Number of executions for each number of neurons
  num.executions <- 10
  
  # Store all number of instances from each class for all executions
  all.num.instances.executions <- list()
  
  # Test many configurations
  repeat{
    
    # Store fmeasure for each execution
    fmeasure.executions <- matrix(nrow=0, ncol=num.evaluation.windows, data=0)
    
    # Number of instances from each class for executions
    num.instances.executions <- matrix(nrow=num.executions, ncol=num.classes, data=0)
    
    for(num.exec in 1:num.executions){
      
      # Was execution successfull?
      executed <- 1
      
      # Store the som maps and micro-clusters
      # [[1]] <- som map
      # [[2]] <- corresponding micro-clusters
      som.micro.clusters <- lapply(1:num.classes, function(x) as.list(rep(0,2)))
      
      # Split training data in 60% training and 40% validation
      ds <- create_holdout_partition(mldrDataTrain, c(train=0.60, test=0.40), "iterative")
      
      # Dataset, attributes and labels indexes
      mldrDataTrain.labels.index <- ds$train$labels$index
      mldrDataTrain.dataset <- ds$train$dataset
      mldrDataTrain.attributesIndexes <- ds$train$attributesIndexes
      mldrDataValid.labels.index <- ds$test$labels$index
      mldrDataValid.dataset <- ds$test$dataset
      mldrDataValid.attributesIndexes <- ds$test$attributesIndexes
      
      # Check for novel classes
      if(novel.classes[1] != 0){
        mldrDataTrain.labels.index <- mldrDataTrain.labels.index[-novel.classes]
        mldrDataValid.labels.index <- mldrDataValid.labels.index[-novel.classes]
      }
      
      # number of initial classes.
      num.classes <- length(mldrDataTrain.labels.index)  
      
      # Separate data from classes
      data.train <- mldrDataTrain.dataset[,mldrDataTrain.attributesIndexes]
      classes.train <- mldrDataTrain.dataset[,mldrDataTrain.labels.index]
      data.valid <- mldrDataValid.dataset[,mldrDataValid.attributesIndexes]
      classes.valid <- mldrDataValid.dataset[,mldrDataValid.labels.index]
      data.train <- as.matrix(data.train)
      data.valid <- as.matrix(data.valid)
      
      # Compute initial class probabilities and class totals
      r <- computeInitialClassProbabilitiesTotals(classes.train)
      class.totals <- r$class.totals
      class.probabilities <- r$class.probabilities
      
      # Store label cardinality of the dataset
      z <- computeLabelCardinality(classes.train)
      
      # Will train a Kohonen map for each class
      for(i in 1:num.classes){
        
        # Execute SOM in data from class i
        data = data.train[which(classes.train[,i]==1),]
        
        # Grid dimension = n.neurons.grid x n.neurons.grid
        mygrid <- somgrid(n.neurons[i],n.neurons[i], topo=c("hexagonal"), neighbourhood.fct = c("gaussian"))
        res <- try(som.map <- som(data, grid=mygrid, rlen = num.iterations, alpha = c(n0, n1), dist.fcts=c("euclidean"), 
                       mode = c(train.mode), maxNA.fraction = 0L, cores=-1), silent=TRUE)#, keep.data = FALSE)
        
        # Verify if execution was successful 
        if(inherits(res, "try-error")){
          #browser()
          executed <- 0
          num.exec <- num.exec - 1
          break;
        }
        
        # Store number of instances for class i in execution num.exec
        num.instances.executions[num.exec,i] <- nrow(data)
        
        # Compute micro-clusters properties and store them
        resultMC <- computeMicroClusters(som.map,i,min.ex)
        
        # We cannot allow only two instances in a neuron! That generates an error!
        if(resultMC$less.than.three == 1){
          less.than.three <- 1
          executed <- 0
          num.exec <- num.exec - 1
          break;
        }
        
        # Get the average output of each neuron based on the mapped instances to them
        average.output.som.map <- getAverageNeuronOutputs(resultMC$som.map,length(resultMC$micro.clusters))
        
        # Calculates conditional probabilitiy thresholds for a class
        # We will consider a p(x|y_j) calculated from the training data (average.output.som.map) for the class for each neuron
        # That means we are considering the outputs of a neuron to be average.output.som.map to calculate the threshold
        # Each neuron will have a threshold
        resultMC$micro.clusters <- getCondProbabilitiesNeurons(resultMC$micro.clusters,class.probabilities,average.output.som.map,i,z)
        
        # Compute radius factor for each neuron
        #resultMC$micro.clusters <- computeRadiusFactorMC(resultMC$micro.clusters,resultMC$som.map,average.output.som.map,data)
        
        # Store the trained SOM
        som.micro.clusters[[i]][[1]] <- resultMC$som.map
        
        # Store the corresponding micro-clusters
        som.micro.clusters[[i]][[2]] <- resultMC$micro.clusters
      }  

      if(executed == 1){
        
        # Test to see the performance of the network in the validation instances
        predictions <- predict.kohonen.win.bayes(som.micro.clusters,data.valid,class.probabilities,z)
        results.valid <- macro.precision.recall.fmeasure.windows(classes.valid,predictions,num.evaluation.windows)
        #results.valid <- macro.precision.recall.fmeasure(classes.valid,predictions)
        fmeasure.executions <- rbind(fmeasure.executions,results.valid$ma.fmeasure.window)
        #cat(length(fmeasure.executions),"\n")
        #cat("\nFmeasure =",results.valid$ma.fmeasure)
      }  
      
    } # for(num.exec in 1:num.executions)
    
    # Store mean fmeasure for current configuration
    mean.fmeasures <- apply(fmeasure.executions,2,mean)
    fmeasures.configurations <- c(fmeasures.configurations,mean(mean.fmeasures))
    
    # Store number of instances for the executions
    all.num.instances.executions[[length(fmeasures.configurations)]] <- num.instances.executions
    
    # Store current configuration
    n.configurations <- n.configurations + 1
    configurations[[n.configurations]] <- n.neurons
    
    cat("\nConfiguration =",n.configurations)
    cat("\nCurrent fmeasure =",fmeasures.configurations[n.configurations])
    cat("\nCurrent configuration =",n.neurons,"\n")
    #prmatrix(num.instances.executions)
    
    # Reduce dimensions of the grids by one neuron
    n.neurons <- n.neurons-1
    
    # If we will not consider novel classes in the stream, we allow maps to have only one neuron
    if(novel.classes[1] == 0){
    
      # If all grids are 0, then stops
      if(sum(n.neurons) == 0){
      #if(sum(n.neurons) == length(n.neurons)){
        break;
      }
      else if(length(n.neurons[n.neurons == 0]) > 0){
      #else if(length(n.neurons[n.neurons == 1]) > 0){
        # Verify if there are empty grids. In this case they are set to 1/2
        n.neurons[which(n.neurons == 0)] <- 1
        #n.neurons[which(n.neurons == 1)] <- 2
      }
    }
    else{
      # If we will consider novel classes, a map must have at least half of its initial number of neurons
      # The smaller the number of neurons, the higher the distance of the weight vector to the farthest instance in the neuron
      # This makes the identification of novel classes much more difficult
      vet.min.neurons <- (n.neurons < ceiling(n.neurons.grid/2))*1
      if(sum(vet.min.neurons) == length(n.neurons)){
        break; # all maps have less than half of their initial number of neurons
      }
      else if(length(vet.min.neurons[vet.min.neurons == 1]) > 0){
        pos.min.neuron <- which(vet.min.neurons==1)
        n.neurons[pos.min.neuron] <- ceiling(n.neurons.grid[pos.min.neuron]/2)
      }
    }
  } # repeat
    
  # Get best configuration
  id.max <- sort(fmeasures.configurations,decreasing=TRUE,index.return=TRUE)$ix[1]
  best.n.neurons <- configurations[[id.max]]
  best.num.instances <- all.num.instances.executions[[id.max]]
  
  results <- list()
  results$best.n.neurons <- best.n.neurons
  results$best.num.instances <- best.num.instances
  
  return(results)
}

# Kohonen map for the offline phase
# -------------------------------------------------------------------------------------
kohonen.offline.multiple <- function(offline.dataset,offline.classes,num.it,init.n,final.n,
                                     n.neurons.grid,tr.mode,min.ex){
  
  cat("\nOffline phase - building maps!")
  
  # Parameters of the algorithm
  n0 <- init.n # initial learning rate
  n1 <- final.n # final learnin rate
  # Learning rate will decline linearly from n0 to n1 over num.iterations updates 
  num.iterations <- num.it # maximum number of iteractions. 
  num.classes <- ncol(offline.classes) # number of initial classes. 
  train.mode <- tr.mode # training mode
  min.ex <- min.ex # minimum number of instances in a neuron
  
  # Compute initial class probabilities and class totals
  r <- computeInitialClassProbabilitiesTotals(offline.classes)
  class.totals <- r$class.totals
  class.probabilities <- r$class.probabilities
  
  # Store label cardinality of the dataset
  z <- computeLabelCardinality(offline.classes)
  
  # Will store the amount of instances in each neuron
  instances.per.neuron <- vector()
  
  # Will store all properties computed here
  result <- list()
  
  # Store the som maps and micro-clusters
  # [[1]] <- som map
  # [[2]] <- corresponding micro-clusters
  som.micro.clusters <- lapply(1:num.classes, function(x) as.list(rep(0,2)))
  
  # We cannot allow only two instances in a neuron! That generates an error!
  less.than.three <- 0
  
  # Will train a Kohonen map for each class
  for(i in 1:num.classes){
    
    # Execute SOM in data from class i
    data = offline.dataset[which(offline.classes[,i]==1),]
    
    # Grid dimension = n.neurons.grid x n.neurons.grid
    mygrid <- somgrid(n.neurons.grid[i],n.neurons.grid[i], topo=c("hexagonal"), neighbourhood.fct = c("gaussian"))
    som.map <- som(data, grid=mygrid, rlen = num.iterations, alpha = c(n0, n1), dist.fcts=c("euclidean"), 
                   mode = c(train.mode), maxNA.fraction = 0L, cores=-1)#, keep.data = FALSE)
    
    # Compute micro-clusters properties and store them
    resultMC <- computeMicroClusters(som.map,i,min.ex)
    
    # We cannot allow only two instances in a neuron! That generates an error!
    if(resultMC$less.than.three == 1){
      less.than.three <- 1
      break;
    }
    
    # Map training data to som.map in order to get p(x|y_j) to calculate conditional probabilitiy threshold
    #output.som.map <- map(resultMC$som.map,data)
    #browser()
    
    # Get the average output of each neuron based on the mapped instances to them
    average.output.som.map <- getAverageNeuronOutputs(resultMC$som.map,length(resultMC$micro.clusters))
    
    # Calculates conditional probabilitiy thresholds for a class
    # We will consider a p(x|y_j) calculated from the training data (average.output.som.map) for the class for each neuron
    # That means we are considering the outputs of a neuron to be average.output.som.map to calculate the threshold
    # Each neuron will have a threshold
    resultMC$micro.clusters <- getCondProbabilitiesNeurons(resultMC$micro.clusters,class.probabilities,average.output.som.map,i,z)

    # Compute radius factor for each neuron
    #resultMC$micro.clusters <- computeRadiusFactorMC(resultMC$micro.clusters,resultMC$som.map,data)
    
    # Store the trained SOM
    som.micro.clusters[[i]][[1]] <- resultMC$som.map
    
    # First index contains the neurons not used
    if(as.numeric(names(table(resultMC$som.map$unit.classif)))[1] == 0){
      instances.per.neuron <- c(instances.per.neuron,table(resultMC$som.map$unit.classif)[-1])
    }
    else{
      instances.per.neuron <- c(instances.per.neuron,table(resultMC$som.map$unit.classif))
    }
    
    # Store the corresponding micro-clusters
    som.micro.clusters[[i]][[2]] <- resultMC$micro.clusters
  }  

  # We cannot allow only two instances in a neuron! That generates an error!
  result$less.than.three <- less.than.three
  
  # Store results
  result$som.micro.clusters <- som.micro.clusters
  result$z <- z
  result$class.probabilities <- class.probabilities
  result$class.totals <- class.totals
  result$total.instances <- nrow(offline.dataset)
  
  return(result)
}
# -------------------------------------------------------------------------------------


# Kohonen map for the online phase considering bayes and label cardinality z
# -------------------------------------------------------------------------------------
kohonen.online.bayes.ND <- function(mapping,online.dataset,init.n,final.n,tr.mode,num.it,
                                    min.ex,novel.classes,n.neurons.grid,
                                    update.model.info,num.offline.instances){

  #plot(1, type="n", xlab="", ylab="", xlim=c(0, 1), ylim=c(0, 1), main=paste(n.neurons.grid," neuronios ",sep=""))
  
	# Time stamp
	time.stamp <- 0

	# Learning rate
	n0 <- init.n
	n1 <- final.n

	# Training mode of the SOM grid
	train.mode <- tr.mode
	
	# Num iterations for Kohonen algorithm
	num.iterations <- num.it

	# Will store results
	results <- list()

	# Mininum number of examples for a valid cluster
	min.examples.cluster <- min.ex

	# Time stamp of the incoming examples
	time.stamp <- 0

	# Will store indexes of explained examples
	indexes.explained <- vector()
	
	# Number of examples in the stream
	num.examples <- nrow(online.dataset)
	
	# Prediction for the known classes
	predictions <- matrix(nrow=0, ncol=nrow(mapping$class.probabilities), data=0)
	colnames(predictions) <- paste("C",c(1:ncol(predictions)),sep="")

	# Number of initial classes
	initial.number.classes <- nrow(mapping$class.probabilities)
	
	# Analyse examples one by one
	cat("\nOnline phase!\n")
	step <- 1
	for(inst in seq(from=1, to=num.examples, by=step)){

	  matrix.examples <- 0
	  if(inst+step > num.examples){
	    matrix.examples <- online.dataset[inst:num.examples,,drop=F]
	    rownames(matrix.examples) <- inst:num.examples
	  }
	  else{
	    matrix.examples <- online.dataset[inst:(inst+(step-1)),,drop=F]
	    rownames(matrix.examples) <- inst:(inst+(step-1))
	  } 

	  rownames.matrix.examples <- as.numeric(rownames(matrix.examples))
	  num.examples.matrix.examples <- nrow(matrix.examples)
	  
	  # Num patterns (sets of neurons) to be considered in the chunck of examples (matrix.examples)
	  num.patterns <- nrow(mapping$class.probabilities)
	  
	  # Get k nearest neighbour neurons for the set of examples
	  winner <- list()
	  all.n.ks <- vector()
	  for(j in 1:num.patterns){
	    #n.k <- ceiling(nrow(mapping$som.micro.clusters[[j]][[1]]$codes[[1]])*0.1)
	    n.k <- nrow(mapping$som.micro.clusters[[j]][[1]]$codes[[1]])
	    if(n.k %% 2 == 0){n.k <- n.k - 1} # Guarantee odd number
	    if(nrow(mapping$som.micro.clusters[[j]][[1]]$codes[[1]])==1){n.k <- 1}
	    winner[[j]] <- get.knnx(mapping$som.micro.clusters[[j]][[1]]$codes[[1]],matrix.examples,k=n.k)
	    all.n.ks <- c(all.n.ks,n.k)
	  }   
	  
	  # For each example
	  for(inst.l in 1:num.examples.matrix.examples){
	    
	    # Time stamp
	    time.stamp <- time.stamp+1
	    
	    # Get incoming example
	    x <- as.numeric(matrix.examples[inst.l,])
	    
	    # Create vector of possible predictions
	    pred <- vector(length=nrow(mapping$class.probabilities), mode="numeric")
	    
	    # Get winning neuron
	    outputs <- vector() 
	    outputs.prob1 <- vector()
	    distances <- vector()
	    winners.index <- vector()
	    
	    # Structures for majority voting
	    vet.distances <- vector()
	    vet.indexes <- vector()
	    
	    # For each class, get the winning neuron among the k closest neurons
	    for(j in 1:num.patterns){
	      
	      # winning neuron
	      win <- winner[[j]]$nn.index[inst.l,1]
	      
	      # Centroid of the winning neuron (weight vector)
	      centroid <- mapping$som.micro.clusters[[j]][[2]][[win]][[1]]
	      
	      # Euclidean distance between centroid of the nearest neuron (micro-cluster) and example
	      distance <- winner[[j]]$nn.dist[inst.l,1]
	      #distance <- euc.dist(centroid,x)
	      
	      # p(x|y_j) considering the winning neuron
	      prob1 <- exp(-distance) 
	      
	      # Class (pos) the neuron represents
	      pos <- mapping$som.micro.clusters[[j]][[2]][[win]][[8]]
	      
	      # p(y_j|x) = p(y_j)*p(x|y_j) considering the winning neuron
	      prob2 <- mapping$class.probabilities[pos,pos]*prob1  
	      
	      winners.index <- c(winners.index,win)
	      distances <- c(distances,distance)
	      outputs.prob1 <- c(outputs.prob1,prob1) 
	      outputs <- c(outputs,prob2)
	      
	      # Storing all distances and indexes of classes
	      # This will be used to decide winner class
	      vet.distances <- c(vet.distances,winner[[j]]$nn.dist[inst.l,])
	      vet.indexes <- c(vet.indexes,rep(j,ncol(winner[[j]]$nn.dist)))
	    }
	    
	    # Sort output values and return sorted indexes
	    prob.sorted <- sort.winning.neurons(distances,vet.distances,vet.indexes,min(all.n.ks),num.patterns,winner,inst.l)
	    idx.prob.sorted <- prob.sorted$ix
	    
	    # Controls if the instance is explained by the model
	    explained <- 0
	    
	    # Check output of the winning neuron
	    radiusFactor1 <- mapping$som.micro.clusters[[idx.prob.sorted[1]]][[2]][[winners.index[idx.prob.sorted[1]]]][[9]]
	    radiusFactor2 <- mapping$som.micro.clusters[[idx.prob.sorted[1]]][[2]][[winners.index[idx.prob.sorted[1]]]][[12]]
	    
	    # Check the max radius for the set of micro-clusters of the winning class
	    #id.class <- idx.prob.sorted[1]
	    #largestDistance <- getLargestDistancesNeuron(mapping$som.micro.clusters,id.class)
	    #radiusFactor1 <- largestDistance[1]
	    #radiusFactor2 <- largestDistance[2]
	    
	    if(novel.classes[1] == 0){
	      radiusFactor1 <- 9999999999999999999999
	      radiusFactor2 <- 9999999999999999999999
	    }   
	    cat(rownames.matrix.examples[inst.l]," - ",outputs.prob1[idx.prob.sorted[1]]," - ",distances[idx.prob.sorted[1]]," - ", radiusFactor1," - ",radiusFactor2,"\n")
      
	    # Check if instance is explained by model
	    if(distances[idx.prob.sorted[1]] <= radiusFactor1){
	      
	      # Instance was explained! 
	      explained <- 1
	      
	      # Ceiling label cardinality 
	      z <- ceiling(mapping$z)
	      for(j in 1:z){
	      #for(j in 1:num.patterns){
	        
	        # Class (pos) the neuron represents
	        pos <- mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[8]]
	        
	        # Average output of the neuron
	        average.output <- mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[11]]

	        if(j == 1){
	          mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[11]] <- (average.output + outputs.prob1[pos])/2
	          pred[pos] <- 1
	        }
	        else{
	          idx.j <- mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[8]]
	          prob.j <- mapping$class.probabilities[idx.j,idx.j] # p(y_j)
	          prob.x.j <- outputs.prob1[idx.prob.sorted[j]] # p(x|y_j)
	          prob.k.j <- 1
	          for(k in 1:(j-1)){ # p(y_k|y_j)
	            idx.k <- mapping$som.micro.clusters[[idx.prob.sorted[k]]][[2]][[winners.index[idx.prob.sorted[k]]]][[8]]
	            if(pred[idx.k] == 1){
	              prob.k.j <- prob.k.j * mapping$class.probabilities[idx.k,idx.j]
	            }
	          }
	          prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
	          
	          # Threshold for neuron winners.index[idx.prob.sorted[j]] considering class idx.j
	          cond.prob.threshold <- mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[10]] 
	          
	          # Test for threshold!
	          if(length(prob.j.ks.x) == 0 | length(cond.prob.threshold) == 0){browser()}
	          if(prob.j.ks.x >= cond.prob.threshold){ 
	            mapping$som.micro.clusters[[idx.prob.sorted[j]]][[2]][[winners.index[idx.prob.sorted[j]]]][[11]] <- (average.output + outputs.prob1[pos])/2
	            pred[pos] <- 1
	          }
	        }
	      } # End for(j in 1:z){
	    } # End if(distances[idx.prob.sorted[1]] <= (std.deviation*radiusFactor1)){
	    
	    if(explained == 1){ # Model explained the example

	      # Store index		
	      indexes.explained <- c(indexes.explained,rownames.matrix.examples[inst.l])
	      
	      # Update information about the model
	      if(update.model.info == 1){
	        mapping <- update.model.information(mapping,pred,x,time.stamp,n0,initial.number.classes,predictions,winner,inst.l)
	      }   
	      
	      # Include predictions in the prediction matrix
	      predictions <- rbind(predictions,pred)
	      rownames(predictions)[nrow(predictions)] <- rownames.matrix.examples[inst.l]
	      
	      # Update class totals and class probabilities
	      pred <- matrix(pred, nrow = 1, ncol = length(pred))
	      mapping <- updateClassTotalsProbabilities(mapping,pred,1,predictions,initial.number.classes,num.offline.instances)
	      
	      # Update label cardinality
	      mapping$z <- ( (mapping$total.instances - 1) * mapping$z + sum(apply(pred,1,sum)) ) / mapping$total.instances
	      
	      # Update conditional probabilitiy thresholds
	      mapping$som.micro.clusters <- updateCondProbabilitiesNeurons(mapping$som.micro.clusters,mapping$class.probabilities,mapping$z)
	      
	    }
	  } # End for(inst.l in 1:num.examples.matrix.examples){
	} # End for(inst in seq(from=1, to=num.examples, by=step))
	
	cat("\n")
	
	results$predictions <- predictions
	results$indexes.explained <- indexes.explained
	results$mapping <- mapping
	
	return(results)
}
# -------------------------------------------------------------------------------------
