##########################################################################
# First implementation of multi-label stream classification with kohonen #
# Ricardo Cerri - 03/05/2019 									    	                     #
##########################################################################

# Auxiliary functions
source("functions.R")
source("kohonen.R")

# Packages to work with multi-label data (mldr_0.4.2) and (utiml_0.1.6)
library(mldr)
library(utiml)

# Package for get.knn function (FNN_1.1.3 )
library(FNN)

# Package for the Kohonen neural network (kohonen_3.0.8)
library(kohonen)

# Command line argument (parameter file)
args <- commandArgs(trailingOnly = TRUE)

# Read parameter values
parameters <- getParameterValues(args[1])
#parameters <- getParameterValues("Config/scene-conf.txt")

# Load datasets
mldrDataTrain <- mldr(parameters$train.data)
mldrDataTest <- mldr(parameters$test.data)

# Get data information
mldrDataTrain.labels.index <- mldrDataTrain$labels$index
mldrDataTrain.dataset <- mldrDataTrain$dataset
mldrDataTrain.attributesIndexes <- mldrDataTrain$attributesIndexes
num.labels <- nrow(mldrDataTrain$labels)
num.classes <- parameters$num.classes # Number of known classes
novel.classes <- 0 # The novelty classes -> no novel classes

# Test Data
mldrDataTest.dataset <- mldrDataTest$dataset
mldrDataTest.attributesIndexes <- mldrDataTest$attributesIndexes
mldrDataTest.labels.index <- mldrDataTest$labels$index

# Get offline dataset
offline.dataset <- mldrDataTrain.dataset[,mldrDataTrain.attributesIndexes]
offline.dataset <- as.matrix(offline.dataset)
if(novel.classes[1] != 0){
  offline.classes <- mldrDataTrain.dataset[,mldrDataTrain.labels.index[-novel.classes]]
  online.classes <- mldrDataTest.dataset[,mldrDataTest.labels.index[-novel.classes]]
}else{
  offline.classes <- mldrDataTrain.dataset[,mldrDataTrain.labels.index]
  online.classes <- mldrDataTest.dataset[,mldrDataTest.labels.index]
}

# Normalize data
#offline.dataset <- scale(offline.dataset)
#offline.dataset <- apply(offline.dataset,2,norm.min.max)

# Get kohonen map information and parameters
n0 <- parameters$n0
n1 <- parameters$n1
tr.mode <- parameters$train.mode
num.iterations <- parameters$num.iterations
min.ex <- 3 # minimum number of instances in a neuron
n.neurons.grid <- parameters$grid.dimension
update.model.info <- parameters$update.model.info
num.evaluation.windows <- parameters$num.evaluation.windows

# Put information about number of neurons
input.neurons.grid <- rep(n.neurons.grid,num.classes)

# Number of instances for the offline phase
num.offline.instances <- nrow(offline.dataset)

# Search for the initial number of neurons that make it possible to execute kohonen without error
n.neurons.grid <- getInitialNeuronGrid(mldrDataTrain,novel.classes,num.iterations,n0,n1,input.neurons.grid,tr.mode)

# Search for the best number of neurons using the training data
results.n.neurons.grid <- searchNumberOfNeurons(mldrDataTrain,num.iterations,n0,n1,n.neurons.grid,tr.mode,min.ex,novel.classes,ncol(offline.classes),num.evaluation.windows)
n.neurons.grid <- results.n.neurons.grid$best.n.neurons
best.num.instances <- results.n.neurons.grid$best.num.instances
#n.neurons.grid <- input.neurons.grid

# These n.neurons.grid values were obtained with 40% of the training data
# Now obtain the number of neurons proportional to the whole offline dataset
#average.num.instances <- apply(best.num.instances,2,mean)

# n.neurons.grid was obtained with 40% of training data
# Let's get the corresponding number for the whole dataset, using a simple rule
#n.neurons.grid <- adjustNeuronsGrid(n.neurons.grid,average.num.instances,offline.classes)

# Execute only once?
only.once <- 1

# Run experiments for different number of neurons (until one neuron per class)
repeat{
  
  # Store results of all executions
  maps <- list()
  vet <- vector()
  vet.NP <- vector()
  
  # Repeat the experiment num.exp times
  executed <- 1
  for(num.exp in 1:parameters$num.experiments){
    
    cat("\nExperiment ",num.exp)
    
    # Build kohonen map in the offline examples
    res <- try(mapping <- kohonen.offline.multiple(offline.dataset,offline.classes,num.iterations,n0,n1,n.neurons.grid,tr.mode,min.ex),silent=TRUE)
    if(inherits(res, "try-error") | mapping$less.than.three == 1){
      executed <- 0
      break
    }
    
    mapping$n.neurons.grid <- n.neurons.grid
    
    # Beginning of the online phase
    
    # Examples without classes
    online.dataset <- mldrDataTest.dataset[,mldrDataTest.attributesIndexes]
    
    # Get online dataset
    online.dataset <- as.matrix(online.dataset)
    
    # Run online Kohonen
    results <- kohonen.online.bayes.ND(mapping,online.dataset,n0,n1,tr.mode,num.iterations,min.ex,novel.classes,n.neurons.grid,
                                       update.model.info,num.offline.instances)
    
    # Evaluate the online predictions. Only considering the known classes
    true.classes <- online.classes[results$indexes.explained,]
    
    results.online <- macro.precision.recall.fmeasure.windows(true.classes,results$predictions,num.evaluation.windows)
    results$true.classes <- true.classes
    results$results.online <- results.online
    
    # Save results of all executions
    maps[[num.exp]] <- results
    vet <- c(vet,results.online$ma.fmeasure)

  } # End for(num.exp in 1:parameters$num.experiments)
  
  # Save the predictions, parameter values and model
  if(executed == 1){
    name.dataset <- unlist(strsplit(parameters$test.data,"/"))
    name.dataset <- gsub(" ","",name.dataset[length(name.dataset)])
    time <- format(Sys.time(), "%H.%M.%OS")
    
    capture.output(parameters,file=paste("Results/",name.dataset,"-",time,"-parameters-",n.neurons.grid[1],".txt",sep=""))
    saveRDS(maps, file=paste("Results/",name.dataset,"-",time,"-model-",n.neurons.grid[1],".rds",sep=""))
    capture.output(
      cat("./Dataset: ",name.dataset,", Mean: ",mean(vet),", StdDev: ",sd(vet)),
      file=paste("Results/",name.dataset,"-",time,"-mean.fmeasures-",n.neurons.grid[1],".txt",sep="")
    )
    #readRDS
  } # End if(executed == 1)
  
  if(only.once == 1){
    break
  }
  
  # Reduce dimensions of the grids by one neuron
  n.neurons.grid <- n.neurons.grid-1
  
  # If all grids are 0, then stops
  if(sum(n.neurons.grid) == 0){
    break
  }
  else if(length(n.neurons.grid[n.neurons.grid == 0]) > 0){
    # Verify if there are empty grids. In this case they are set to 1
    n.neurons.grid[which(n.neurons.grid == 0)] <- 1
  }
}
# -------------------------------------------------------------------------------------


